/*
								+----------------------------------+
								|                                  |
								| ***   Mersene twister rand   *** |
								|                                  |
				http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/MT2002/CODES/mt19937ar.c
								|                                  |
								|             Mersene.h            |
								|                                  |
								+----------------------------------+
*/

/*
 *	2007-11-14
 *
 *	enhanced linux compatibility by adding unix integer types
 *
 *	2008-03-04
 *
 *	now using Integer.h header
 *
 */

#ifndef __MERSENE_INCLUDED
#define __MERSENE_INCLUDED

#include "Integer.h"

class CMerseneTwister {
protected:
	enum {
		N = 624,
		M = 397,
		MATRIX_A = 0x9908b0dfUL,  /* constant vector a */
		UPPER_MASK = 0x80000000UL, /* most significant w-r bits */
		LOWER_MASK = 0x7fffffffUL /* least significant r bits */
	};

	unsigned long m_n_state[N]; /* the array for the state vector  */
	int m_n_init_index; /* m_n_init_index==N+1 means m_n_state[N] is not initialized */

public:
	CMerseneTwister();

	/* initializes m_n_state[N] with a seed */
	void init_genrand(unsigned long s);

	/* initialize by an array with array-length */
	/* init_key is the array for initializing keys */
	/* key_length is its length */
	/* slight change for C++, 2004/2/26 */
	void init_by_array(unsigned long init_key[], int key_length);

	/* generates a random number on [0,0xffffffff]-interval */
	unsigned long genrand_int32(void);

	/* generates a random number on [0,0x7fffffff]-interval */
	long genrand_int31(void);

	/* generates a random number on [0,1]-real-interval */
	double genrand_real1(void);

	/* generates a random number on [0,1)-real-interval */
	double genrand_real2(void);

	/* generates a random number on (0,1)-real-interval */
	double genrand_real3(void);

	/* generates a random number on [0,1) with 53-bit resolution*/
	double genrand_res53(void);
};

struct TMerseneCache {
protected:
	enum {
		cache_Length = 256, // length of queue
		cache_Paragraph = 8 // number of queries to generate
	};

	struct TCacheItem {
		uint32_t n_value;
		uint64_t n_index;
	} m_t_cache[cache_Length];
	int m_n_cache_start; // cache is basically round buffer, overwriting itself again and again

	unsigned long m_n_twister_init[4];
	uint64_t m_n_twister_position;
	CMerseneTwister m_mt;

public:
	TMerseneCache(unsigned long n_init_array[4]);

	// return aligned value, n_val_index is index of byte-quad, not index of first byte
	uint32_t n_AlignedValue(uint64_t n_val_index);

	// return value n_val_index with any alignment (index is now in bytes)
	inline uint32_t n_Rand32(uint64_t n_val_index)
	{
		if(!(n_val_index % 4))
			return n_AlignedValue(n_val_index / 4);
		int n_bits_a = (int)(n_val_index % 4) * 8;
		n_val_index /= 4;
		return (n_AlignedValue(n_val_index) << n_bits_a) |
			(n_AlignedValue(n_val_index + 1) >> (32 - n_bits_a));
	}

	// return value n_val_index with any alignment (index is now in bytes)
	inline uint16_t n_Rand16(uint64_t n_val_index)
	{
		if(n_val_index % 4 < 3) {
			return (uint16_t)((n_AlignedValue(n_val_index / 4) >>
				(16 - (int)(n_val_index % 4) * 8)) & 0xffff);
		}
		n_val_index /= 4;
		return (uint16_t)(((n_AlignedValue(n_val_index) << 8) & 0xff00) |
			((n_AlignedValue(n_val_index + 1) >> 24) & 0xff));
	}

	// return value n_val_index with any alignment (index is now in bytes)
	inline unsigned char n_Rand8(uint64_t n_val_index)
	{
		return (unsigned char)((n_AlignedValue(n_val_index / 4) >>
			(24 - (int)(n_val_index % 4) * 8)) & 0xff);
	}

protected:
	/*
	 *	void Cache_Swap(int n_index_a, int n_index_b)
	 *		- swap two items in cache, so the most wanted item is not likely to be removed from cache
	 */
	void Cache_Swap(int n_index_a, int n_index_b);
};

#endif //__MERSENE_INCLUDED
