/*
								+--------------------------------+
								|                                |
								|  ***   Simple threading   ***  |
								|                                |
								|  Copyright  -tHE SWINe- 2008  |
								|                                |
								|            Thread.h            |
								|                                |
								+--------------------------------+
*/

#ifndef __LAME_THREADS_INCLUDED
#define __LAME_THREADS_INCLUDED

/*
 *	class CRunnable
 *		- virtual class for runnable object
 */
class CRunnable {
public:
	virtual void Run() = 0;
};

#ifdef WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif

/*
 *	class CThread
 *		- simple thread class; todo - create linux version
 */
class CThread {
protected:
	CRunnable *m_p_runnable;

#ifdef WIN32
	HANDLE m_h_thread;
#else
	pthread_t m_t_thread; // thread
	pthread_mutex_t t_running_mutex;
	bool m_b_running; // thread state
#endif

	class CEmptyRunnable : public CRunnable {
		virtual void Run() {}
	};
	static CEmptyRunnable empty_runnable;

public:
	/*
	 *	CThread::CThread()
	 *		- default constructor, attaches empty runnable object to a thread
	 */
	CThread();

	/*
	 *	CThread::CThread(CRunnable &r_runnable)
	 *		- constructor, attaches runnable object to a thread
	 */
	CThread(CRunnable &r_runnable);

	/*
	 *	CThread::~CThread()
	 *		- destructor
	 */
	~CThread();

	/*
	 *	bool CThread::AttachRunnable(CRunnable &r_runnable)
	 *		- attaches different runnable, thread must not be running
	 *		- returns true on success, false on failure
	 */
	bool AttachRunnable(CRunnable &r_runnable);

	/*
	 *	CRunnable &CThread::r_Runnable()
	 *		- returns reference to runnable object
	 */
	CRunnable &r_Runnable();

	/*
	 *	const CRunnable &CThread::r_Runnable() const
	 *		- returns const reference to runnable object
	 */
	const CRunnable &r_Runnable() const;

	/*
	 *	bool CThread::Start()
	 *		- starts the thread and returns immediately
	 *		- note this fails in case the thread is already running
	 *		- returns true on success, false on failure
	 */
	bool Start();

	/*
	 *	bool CThread::Run()
	 *		- starts the thread and waits untill it finishes
	 *		- note this fails in case the thread is already running
	 *		- also note this is not thread-safe (the Thread object doesn't change
	 *		  it's state to running while in this function and also can't be stopped)
	 *		- returns true on success, false on failure
	 */
	bool Run();

	/*
	 *	bool CThread::b_IsRunning()
	 *		- returns true if thread is running, otherwise returns false
	 */
	bool b_IsRunning();

	/*
	 *	bool CThread::Stop(bool b_force_kill = false)
	 *		- waits for the thread to end, if b_force_kill is set, it's ended forcibly
	 *		- returns true if thread is stopped, returns false if thread didn't
	 *		  stop in given time or on failure
	 */
	bool Stop(bool b_force_kill = false);

	/*
	 *	static int CThread::n_CPU_Num();
	 *		- utility function; returns number of CPU's
	 *		- in case CPU count cannot be determined, returns -1
	 */
	static int n_CPU_Num();

protected:
#ifdef WIN32
	static unsigned long __stdcall _run(void *p_arg); // run for windows
#else
	static void *_run(void *p_arg); // run for linux
#endif
	const CThread &operator =(const CThread &r_thread) { return r_thread; } // can't copy threads this way
};

#endif //__LAME_THREADS_INCLUDED
